import threading
from lox import RWLock
from copy import copy
from time import sleep, time
from collections import deque

SLEEP_TIME = 0.01
N_WORKERS = 5

rw_lock = None
resource = None
resp = None


def common_setup():
    global rw_lock, resp
    rw_lock = RWLock()
    resp = deque()


def common_create_workers(func, n, *args):
    threads = []
    for i in range(n):
        t = threading.Thread(target=func, args=args)
        threads.append(t)

    t_start = time()
    for t in threads:
        t.start()

    return threads, t_start


def read_worker():
    global rw_lock, resp
    with rw_lock('r'):
        local_copy = copy(resource)
        sleep(SLEEP_TIME)

    resp.append(local_copy)
    return


def write_worker(val):
    global rw_lock, resource
    with rw_lock('w'):
        resource = val
    return


def test_RWLock_r():
    global rw_lock, resource, resp
    common_setup()
    resource = 0

    threads, t_start = common_create_workers(read_worker, N_WORKERS)
    for t in threads:
        t.join()
    t_end = time()
    t_diff = t_end-t_start

    assert(N_WORKERS > 2)
    # for this to be true, readers have to access at same time (good)
    assert(t_diff < (N_WORKERS-1)*SLEEP_TIME)
    for r in resp:
        assert(r == resource)


def test_RWLock_w():
    global rw_lock, resource
    common_setup()
    resource = 0
    new_val = 5

    threads_w1, t_start_w1 = common_create_workers(write_worker, 1, new_val)

    for t in threads_w1:
        t.join()

    assert(resource == new_val)


def test_RWLock_rw():
    global rw_lock, resource, resp
    common_setup()
    return
    resource = 0
    soln = [0, ]*N_WORKERS + [5, ]*N_WORKERS

    threads_r1, t_start_r1 = common_create_workers(read_worker, N_WORKERS)
    threads_w1, t_start_w1 = common_create_workers(write_worker, N_WORKERS, 5)
    threads_r2, t_start_r2 = common_create_workers(read_worker, N_WORKERS)

    for t in threads_r1:
        t.join()
    for t in threads_w1:
        t.join()
    for t in threads_r2:
        t.join()

    for r, s in zip(resp, soln):
        assert(r == s)


def test_RWLock_timeout():
    lock = RWLock()

    assert(lock.acquire('r', timeout=0.01) is True)
    assert(lock.acquire('w', timeout=0.01) is False)
    assert(lock.acquire('r', timeout=0.01) is True)
    lock.release('r')
    lock.release('r')

    assert(lock.acquire('w', timeout=0.01) is True)
    assert(lock.acquire('w', timeout=0.01) is False)
    assert(lock.acquire('r', timeout=0.01) is False)
    lock.release('w')


def test_bathroom_example():
    # Note: after the janitor exits, the remaining people are nondeterministic
    sol = [
        "p_0_enter",
        "p_0_exit",
        "j_enter",
        "j_exit",
    ]

    res = bathroom_example()[:4]

    for r, s in zip(res, sol):
        assert(r == s)


def bathroom_example():
    """
    Scenario:
        A janitor needs to clean a restroom, but is not allowed to enter until
        all people are out of the restroom. How do we implement this?
    """
    restroom = RWLock()
    res = []
    n_people = 5
    sleep_time = 0.1

    def janitor():
        with restroom('w'):  # block until the restroom is no longer occupied
            res.append('j_enter')
            print("(%0.3f s) Janitor  entered the restroom" % (time() - t_start,))
            sleep(sleep_time)  # clean the restroom
            res.append('j_exit')
            print("(%0.3f s) Janitor  exited  the restroom" % (time() - t_start,))

    def people(id):
        if id == 0:  # Get the starting time of execution for display purposes
            global t_start
            t_start = time()
        with restroom('r'):  # block if a janitor is in the restroom
            res.append("p_%d_enter" % (id,))
            print("(%0.3f s) Person %d entered the restroom" % (time() - t_start, id,))
            sleep(sleep_time)  # use the restroom
            res.append("p_%d_exit" % (id,))
            print("(%0.3f s) Person %d exited  the restroom" % (time() - t_start, id,))

    people_threads = [threading.Thread(target=people, args=(i,)) for i in range(n_people)]
    janitor_thread = threading.Thread(target=janitor)

    for i, person in enumerate(people_threads):
        person.start()                 # Person i will now attempt to enter the restroom
        sleep(sleep_time * 0.6)        # wait for 60% the time a person spends in the restroom
        if i == 0:                       # While the first person is in the restroom...
            janitor_thread.start()     # the janitor would like to enter. HOWEVER...
            # A new person (until all n_people are done) enters every 0.5 seconds.
    # Wait for all threads to finish
    for t in people_threads:
        t.join()
    janitor_thread.join()

    # The results will look like:
    """
    Running Restroom Demo
    (0.000 s) Person 0 entered the restroom
    (0.100 s) Person 0 exited  the restroom
    (0.101 s) Janitor  entered the restroom
    (0.201 s) Janitor  exited  the restroom
    (0.201 s) Person 1 entered the restroom
    (0.202 s) Person 2 entered the restroom
    (0.202 s) Person 3 entered the restroom
    (0.243 s) Person 4 entered the restroom
    (0.302 s) Person 1 exited  the restroom
    (0.302 s) Person 2 exited  the restroom
    (0.303 s) Person 3 exited  the restroom
    (0.343 s) Person 4 exited  the restroom
    """
    # While Person 0 is in the restroom, the Janitor is waiting to enter (at around 0.5000 s).
    # While the Janitor is waiting, he doesn't let anyone else into the room.
    # After Person 0, leaves the room, the Janitor enters.
    # After cleaning, the Janitor leaves at the 2.000 second mark.
    # Ever since the janitor was waiting (at 0.500 s), Person 1, Person 2,
    # Person 3, and Person 4 have been lining up to enter.
    # Now that the Janitor left the restroom, all the waiting people go in at the same time.
    return res


if __name__ == "__main__":
    print("Running Restroom Demo")
    bathroom_example()



"""
.. module:: rw_Lock
   :synopsis: Synchronization primitive to solves the readers–writers problem.

.. moduleauthor:: Brian Pugh <bnp117@gmail.com>
"""

from threading import Lock
from contextlib import contextmanager
from . import LightSwitch

__all__ = ["RWLock", ]


class RWLock:
    """Lock for a Multi-Reader-Single-Writer scenario.

    Unlimited numbers of reader can obtain the lock, but as soon as a writer
    attempts to acquire the lock, all reads are blocked until the current
    readers are finished, the writer acquires the lock, and finally 
    releases it.

    Similar to a ``lox.LightSwitch``, but blocks incoming "readers" while a "write"
    is trying to be performed.

    Attributes
    ----------
    read_counter : int
        Number of readers that have acquired the lock.
    """

    def __init__(self):
        """ Create RWLock object """

        self._no_writers = Lock()
        self._no_readers = Lock()
        self.read_counter = LightSwitch(self._no_writers)
        self._write_counter = LightSwitch(self._no_readers)
        self._readers_queue = Lock()

    @contextmanager
    def __call__(self, rw_flag: str, timeout=-1):
        """ Used in contextmanager to specify acquire/release type """

        self.acquire(rw_flag, timeout=timeout)
        try:
            yield self
        finally:
            self.release(rw_flag)

    def __len__(self):
        """Get the read_counter value

        Returns
        -------
        int
           Number of current readers
        """

        return len(self.read_counter)

    def _check_rw_flag(self, rw_flag):
        """Checks if passed in flag is a valid value"""

        rw_flag = rw_flag.lower()
        if rw_flag == 'r':
            pass
        elif rw_flag == 'w':
            pass
        else:
            raise ValueError("rw_flag must be 'r' or 'w'")
        return rw_flag

    def acquire(self, rw_flag: str, timeout=-1):
        """ Acquire the lock as a "reader" or a "writer".

        Parameters
        ----------
        rw_flag : str
            Either 'r' for 'read' or 'w' for 'write' acquire.

        timeout : float
            Time in seconds before timeout occurs for acquiring lock.

        Returns
        -------
        bool
            ``True`` if lock was acquired, ``False`` otherwise.
        """

        obtained = False
        rw_flag = self._check_rw_flag(rw_flag)
        if rw_flag == 'r':
            obtained = self._readers_queue.acquire(timeout=timeout)
            if not obtained:
                return False

            obtained = self._no_readers.acquire(timeout=timeout)
            if not obtained:
                self._readers_queue.release()
                return False

            obtained = self.read_counter.acquire(timeout=timeout)

            self._no_readers.release()
            self._readers_queue.release()
        elif rw_flag == 'w':
            obtained = self._write_counter.acquire(timeout=timeout)
            if not obtained:
                return False

            obtained = self._no_writers.acquire(timeout=timeout)
            if not obtained:
                self._write_counter.release()
        return obtained

    def release(self, rw_flag: str):
        """Release acquired lock.

        Parameters
        ----------
        rw_flag : str
            Either 'r' for 'read' or 'w' for 'write' acquire.
        """

        rw_flag = self._check_rw_flag(rw_flag)
        if rw_flag == 'r':
            self.read_counter.release()
        elif rw_flag == 'w':
            self._no_writers.release()
            self._write_counter.release()



import lox
from lox.helper import term_colors, cdf, cdfi
from time import sleep, time
import sys
import gc

SLEEP_TIME = 0.01


def test_ctx(capsys):
    with capsys.disabled():
        print("")
        with term_colors("red"):
            print("test red")
        with term_colors("green"):
            print("test green")
        with term_colors("blue"):
            print("test blue")
        with term_colors("orange"):
            print("test orange")
        with term_colors("underline"):
            print("test underline")
        with term_colors("bold"):
            print("test bold")
        with term_colors("header"):
            print("test header")


class Class1:

    def __init__(self, z):
        self.z = z

    @lox.thread(2)
    def test_method1(self, x, y):
        sleep(SLEEP_TIME)
        return x * y + self.z

    @lox.thread
    def test_method2(self, x, y):
        sleep(SLEEP_TIME)
        return x * y - self.z


def test_MethodDecoratorAdaptor_gc():
    len_cdfi = len(cdfi)

    for i in range(100):
        obj = Class1(i)
        obj.test_method1.scatter(i, i+1)
        obj.test_method1.gather()
        del obj
    gc.collect()

    # cdfi should be as it previously was, otherwise gc isn't working
    assert(len_cdfi == len(cdfi))



from threading import Lock, Thread
from lox import QLock
from time import time, sleep

SLEEP_TIME = 0.1


def test_1():
    res = ""
    sol = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    qlock = QLock()

    def worker(x):
        nonlocal res
        with qlock:
            res += x

    threads = []
    for name in sol:
        t = Thread(target=worker, args=(name,))
        t.start()
        threads.append(t)
        sleep(SLEEP_TIME)  # probably enough time for .acquire() to run

    # Wait for all threads to complete
    for t in threads:
        t.join()

    for r, s in zip(res, sol):
        assert r == s


def test_timeout():
    qlock = QLock()
    assert(qlock.acquire())
    assert(qlock.acquire(timeout=SLEEP_TIME) is False)


def test_perf_qlock(benchmark):
    lock = QLock()

    @benchmark
    def acquire_release():
        lock.acquire()
        lock.release()


def test_perf_lock(benchmark):
    lock = Lock()

    @benchmark
    def acquire_release():
        lock.acquire()
        lock.release()



from abc import ABC, abstractmethod
from collections import deque
import threading
from ..exceptions import Timeout

from threading import Lock
import pathos.multiprocessing as mp

__all__ = ["WorkerWrapper", "ScatterPromise"]


class WorkerWrapper(ABC):
    """ Worker helper decorator.

    Methods
    -------
    __call__( *args, **kwargs )
        Vanilla passthrough function execution. Default user function behavior.

    __len__()
        Returns the current job queue length

    scatter( *args, **kwargs)
        Start a job executing func( *args, **kwargs ).
        Workers are spun up automatically.
        Obtain results via gather()

    gather()
        Block until all jobs called via scatter() are complete.
        Returns a list of results in the order that scatter was invoked.
    """

    def __init__(self, n_workers, func):
        # Maximum number of workers to spin up
        self.n_workers = n_workers

        # Function metadata
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__
        self.__module__ = func.__module__

        self.response = deque()  # Stores gather'd user function responses

    def __call__(self, *args, **kwargs):
        """ Vanilla execute the wrapped function. """

        return self.func(*args, **kwargs)

    @abstractmethod
    def __len__(self):
        """
        Returns
        -------
        int
            Length of unprocessed job queue.
        """

        return

    @abstractmethod
    def scatter(self, *args, **kwargs):
        """Enqueue a job to be processed by workers.
        Spin up workers if necessary.

        Returns
        -------
        int-like
            Index into solution array. May be a int-like promise for chaining.
        """
        return

    @abstractmethod
    def gather(self):
        """ Gather results. Blocks until job_queue is empty.
        Also blocks scatter on other threads.

        Returns
        -------
        list
            Results in order ``scatter``'d
        """
        return


class ScatterPromise(int):
    """ Represents: index into solution array.
    Also provides the functionality to chain scatter calls.
    """

    def __new__(cls, val, dec):
        """
        Parameters
        ----------
        val : int
            Index of result into solution array.
        dec : decorator object
            Decorator object.
        """

        if val < 0:
            raise ValueError("Value must be a non-negative index")
        new_obj = super(cls, cls).__new__(cls, val)
        new_obj.dec = dec  # Decorator object
        return new_obj



import threading
import lox
from time import sleep, time
import logging as log

SLEEP_TIME = 0.01
N_WORKERS = 5


def test_basic_args():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.thread(N_WORKERS)  # specifying maximum number of threads
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert(worker_task(2, 5) == 10)

    assert(len(worker_task) == 0)

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y) == r)


def test_basic_noargs():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.thread  # defaults to a max of 50 threads if none specified
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert(10 == worker_task(2, 5))

    assert(len(worker_task) == 0)

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)
    res = worker_task.gather()

    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y) == r)


class Class1:

    def __init__(self, z):
        self.z = z

    @lox.thread(2)
    def test_method1(self, x, y):
        sleep(SLEEP_TIME)
        return x*y + self.z

    @lox.thread
    def test_method2(self, x, y):
        sleep(SLEEP_TIME)
        return x*y - self.z


def test_method_1():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]
    z1 = 5
    z2 = 5

    test_obj_1 = Class1(z1)
    test_obj_2 = Class1(z2)

    assert((2*5+z1) == test_obj_1.test_method1(2, 5))
    assert((2*5-z1) == test_obj_1.test_method2(2, 5))

    assert((2*5+z2) == test_obj_2.test_method1(2, 5))
    assert((2*5-z2) == test_obj_2.test_method2(2, 5))

    for i in range(2):
        for x, y in zip(in_x, in_y):
            test_obj_1.test_method1.scatter(x, y)
        res = test_obj_1.test_method1.gather()
        assert(len(res) == len(in_x))
        for r, x, y in zip(res, in_x, in_y):
            assert((x*y+z1) == r)

    for x, y in zip(in_x, in_y):
        test_obj_2.test_method1.scatter(x, y)
    res = test_obj_2.test_method1.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert((x*y+z2) == r)

    for x, y in zip(in_x, in_y):
        test_obj_1.test_method2.scatter(x, y)
    res = test_obj_1.test_method2.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert((x*y-z1) == r)

    for x, y in zip(in_x, in_y):
        test_obj_2.test_method2.scatter(x, y)
    res = test_obj_2.test_method2.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert((x*y-z2) == r)


def test_chaining_1():
    """ Naive chaining with no additional arguments
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.thread(2)
    def foo(x, y):
        return x, x*y

    @lox.thread(2)
    def bar(x, y):
        return x + y

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y))
    res = bar.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(x+(x*y) == r)

    log.info("First chain call successful")

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(y, x))
    res = bar.gather()  # Should internally call gather of all previous functions
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(y+(y*x) == r)


def test_chaining_1_1():
    """ Naive chaining when func doesn't return a tuple
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.thread(2)
    def foo(x, y):
        return x + y

    @lox.thread(2)
    def bar(x):
        return x ** 2

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y))
    res = bar.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert((x+y)**2 == r)

    log.info("First chain call successful")

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y))
    res = bar.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert((x+y)**2 == r)


def test_chaining_1_2():
    """ Naive chaining when func return a tuple without auto_unpacking.
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.thread(2)
    def foo(x, y):
        return x**2, y**2

    @lox.thread(2)
    def bar(x):
        return x[0] + x[1]

    bar.disable_auto_unpacking()

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y))
    res = bar.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(x**2+y**2 == r)

    log.info("First chain call successful")

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y))
    res = bar.gather()
    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(x**2+y**2 == r)


def test_chaining_2():
    """ Test positional arguments while chaining
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]
    z = 5

    @lox.thread(2)
    def foo(x, y):
        return x, x*y

    @lox.thread(2)
    def bar(x, y, z):
        return x + y + 2*z

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y), z)
    res = bar.gather()

    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(x+(x*y)+2*z == r)

    for x, y in zip(in_x, in_y):
        bar.scatter(z, foo.scatter(x, y))
    res = bar.gather()

    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(bar(z, *foo(x, y)) == r)


def test_chaining_3():
    """ Test kwargs arguments while chaining
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]
    z = 5

    @lox.thread(2)
    def foo(x, y):
        return x, x*y

    @lox.thread(2)
    def bar(x, y, dummy=0, z=-1):
        return x + y + 2*z

    for x, y in zip(in_x, in_y):
        bar.scatter(foo.scatter(x, y), z=z)
    res = bar.gather()

    assert(len(res) == len(in_x))
    for r, x, y in zip(res, in_x, in_y):
        assert(x+(x*y)+2*z == r)


def test_chaining_4():
    """ Test parallel chaining
    """

    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]
    z = 5

    @lox.thread(2)
    def foo(x, y):
        return x, x*y

    @lox.thread(2)
    def bar(x, y, dummy=0, z=-1):
        return x + y + 2*z

    @lox.thread(2)
    def baz(a, b):
        return a*b

    for x, y in zip(in_x, in_y):
        res1 = foo.scatter(x, y)
        bar.scatter(res1, z=z)
        baz.scatter(res1)
    bar_res = bar.gather()
    baz_res = baz.gather()

    assert(len(bar_res) == len(in_x))
    assert(len(baz_res) == len(in_x))

    for r, x, y in zip(bar_res, in_x, in_y):
        assert(bar(*foo(x, y), z=z) == r)
    for r, x, y in zip(baz_res, in_x, in_y):
        assert(baz(*foo(x, y)) == r)



"""
.. module:: LightSwitch
   :synopsis: Lock acquire on first entry and release on last exit.

.. moduleauthor:: Brian Pugh <bnp117@gmail.com>

See https://w3.cs.jmu.edu/kirkpams/OpenCSF/Books/cs361/html/DesignAdv.html

"""

import threading
import pathos.multiprocessing as mp
from time import time

__all__ = ["LightSwitch", ]


class LightSwitch:
    """Acquires a provided lock while ``LightSwitch`` is in use.

    The lightswitch pattern creates a first-in-last-out synchronization
    mechanism. The name of the pattern is inspired by people entering a
    room in the physical world. The first person to enter the room turns
    on the lights; then, when everyone is leaving, the last person to exit
    turns the lights off.

    Attributes
    ----------
    lock : threading.Lock
        The lock provided to the constructor that may be acquired/released
        by ``LightSwitch``.

    counter: int
        Number of times the ``LightSwitch`` has been acquired without release.
    """

    def __init__(self, lock, multiprocessing=False):
        """Create a ``LightSwitch`` object.

        Parameters
        ----------
        lock : ``threading.Lock`` or ``pathos.multiprocessing.Lock``
            Lock to acquire when internal counter is incremented from zero.
            Lock to release when internal counter is decremented to zero.
        """

        if multiprocessing:
            self._library = mp
        else:
            self._library = threading

        self.lock = lock
        self.counter = 0
        self._counter_lock = self._library.Lock()

    @property
    def lock(self):
        return self._lock

    @lock.setter
    def lock(self, lock):
        self._lock = lock

    @property
    def counter(self):
        return self._counter

    @counter.setter
    def counter(self, counter):
        self._counter = counter

    def __enter__(self):
        """ Acquire ``LightSwitch`` at context enter. """

        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Release ``LightSwitch`` at context exit. """

        self.release()

    def __len__(self):
        """Get the counter value.

        Returns
        -------
        int
            counter value (number of times lightswitch has been acquired).
        """

        return self.counter

    def acquire(self, timeout=-1):
        """Acquire the ``LightSwitch`` and increment the internal counter.

        When the internal counter is incremented from zero, it will acquire
        the provided lock.

        Parameters
        ----------
        timeout : float
            Maximum number of seconds to wait before aborting.

        Returns
        -------
        bool
            ``True`` on success, ``False`` on failure (like timeout).
        """

        # Acquire the counter_lock while keeping track of approximately time
        t_start = time()
        if not self._counter_lock.acquire(timeout=timeout):
            return False
        t_remaining = timeout - (time() - t_start)

        if timeout < 0:
            t_remaining = -1
        elif t_remaining < 0:
            t_remaining = 0

        # Acquire lock if first to acquire
        if self.counter == 0:
            if not self._lock.acquire(timeout=t_remaining):
                self._counter_lock.release()
                return False
        self.counter += 1

        self._counter_lock.release()
        return True

    def release(self):
        """Release the ``LightSwitch`` by decrementing the internal counter.

        When the internal counter is decremented to zero, it will release
        the provided lock.
        """

        with self._counter_lock:
            self.counter -= 1
            if self.counter == 0:
                self._lock.release()



"""
.. module:: QLock
   :synopsis: FIFO Lock

.. moduleauthor:: Brian Pugh <bnp117@gmail.com>
"""

from threading import Lock
from collections import deque

__all__ = ["QLock", ]


class QLock:
    """Lock that guarentees FIFO operation. Approximately 6x slower than a 
    normal ``Lock()``.

    Modified from https://stackoverflow.com/a/19695878
    """

    def __init__(self):
        """ Create a ``QLock`` object. """

        self.queue = deque()
        self.lock = Lock()
        self.count = 0

    def __enter__(self):
        """ Acquire ``QLock`` at context enter. """

        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Release ``QLock`` at context exit. """

        self.release()

    def __len__(self,):
        """ 
        Returns
        -------
        int
            Number of tasks waiting to acquire.
        """

        return self.count

    @property
    def locked(self):
        """ Whether or not the ``QLock`` is acquired """

        return self.count > 0

    def acquire(self, timeout=-1):
        """Block until resource is available.

        Threads that call ``acquire`` obtain resource FIFO.

        Parameters
        ----------
        timeout : float
            Maximum number of seconds to wait before aborting.

        Return
        ------
        bool
            True on successful acquire, False on timeout.
        """

        with self.lock:
            if self.count > 0:
                # Append a new acquired lock the queue
                lock = Lock()
                lock.acquire()
                self.queue.append(lock)
                self.lock.release()

                # Block until the lock has been released from another thread
                acquired = lock.acquire(timeout=timeout)
                self.lock.acquire()
                if acquired:
                    # lock acquired
                    pass
                else:
                    # timed out
                    try:
                        self.queue.remove(self.lock)
                    except ValueError:
                        # lock must have been released between the timeout
                        # and self.lock.acquire()
                        pass
                    finally:
                        return False
            self.count += 1
        return True

    def release(self):
        """ Release exclusive access to resource.

        Exception
        ---------
        ValueError
            Lock released more than it has been acquired.
        """

        with self.lock:
            if self.count == 0:
                raise ValueError("lock not acquired")
            self.count -= 1
            if self.queue:
                self.queue.popleft().release()



"""
.. module:: index_semaphore
   :synopsis: BoundedSemaphore where acquires return an index from [0, val).

.. moduleauthor:: Brian Pugh <bnp117@gmail.com>
"""

from queue import Queue, Empty, Full
from contextlib import contextmanager
import logging as log

__all__ = ["IndexSemaphore", ]


class IndexSemaphore:
    """ ``BoundedSemaphore``-like object where acquires return an index from [0, val).

    Example usecase: thread acquiring a GPU.

    Example:
        >>> sem = IndexSemaphore(4)
        >>> with sem() as index:
        >>>     print("Obtained resource %d" % (index,))
        >>> 
        Obtained resource 0
    """

    def __init__(self, val):
        """Create an ``IndexSemaphore`` object.

        Parameters
        ----------
        val : int
            Number of resources available.
        """

        if val <= 0:
            raise ValueError("val must be >0")

        self.queue = Queue(maxsize=val)
        for i in range(val):
            self.queue.put(i)

    @contextmanager
    def __call__(self, timeout=None):
        """Enter context manager when a timeout wants to be specified.
        Only to be call as part of a "with" statement.

        Parameters
        ----------
        timeout : float
            Maximum number of seconds to wait before aborting.
            Set timeout=None for no timeout.

        Returns
        -------
        int
            Resource index. None on timeout/failure.
        """

        index = self.acquire(timeout=timeout)
        try:
            yield index
        except Exception as e:
            log.error(e)
        finally:
            if index is not None:
                self.release(index)

    def __len__(self):
        """ 
        Returns
        -------
        int
            Current blocked queue size.
        """

        return self.queue.qsize()

    def acquire(self, timeout=None):
        """Blocking acquire resource.

        Parameters
        ----------
        timeout : float
            Maximum number of seconds to wait before returning.

        Returns
        -------
        int
            Resource index on successful acquire. None on timeout.
        """

        try:
            return self.queue.get(timeout=timeout)
        except Empty:
            return None

    def release(self, index):
        """ Release resource at index.

        Parameters
        ----------
        index : int
            Index of resource to release.

        Raises
        ------
        Exception
            Resource has been released more times than acquired.
        """

        try:
            self.queue.put_nowait(index)
        except Full:
            raise Exception("IndexSemaphore released more times than acquired")



# -*- coding: utf-8 -*-

"""
Concurrency control objects to help parallelized tasks communicate and share resources.
"""

from .light_switch import *
from .rw_lock import *
from .index_semaphore import *
from .qlock import *



import lox
from lox import Announcement
from threading import Lock, Thread
from time import sleep
import queue


def test_1():
    """ Test most basic usecase
    """

    ann = lox.Announcement()
    x_in = [1, 2, 3, 4, 5]
    foo_soln, bar_soln = [], []

    foo_q = ann.subscribe()
    bar_q = ann.subscribe()

    assert(isinstance(foo_q, lox.Announcement))
    assert(isinstance(bar_q, lox.Announcement))
    assert(foo_q != bar_q)

    def foo():
        x = foo_q.get()
        foo_soln.append(x**2)

    def bar():
        x = bar_q.get()
        bar_soln.append(x**3)

    threads = []
    for _ in x_in:
        threads.append(Thread(target=foo))
        threads.append(Thread(target=bar))

    for x in x_in:
        ann.put(x)

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert(len(foo_soln) == len(x_in))
    assert(len(bar_soln) == len(x_in))

    for x, r in zip(x_in, foo_soln):
        assert(r == x**2)

    for x, r in zip(x_in, bar_soln):
        assert(r == x**3)


def test_2():
    """ Test the backlog
    """

    ann = lox.Announcement(backlog=0)
    x_in = [1, 2, 3, 4, 5]
    foo_soln, bar_soln = [], []

    foo_q = ann.subscribe()

    def foo():
        x = foo_q.get()
        foo_soln.append(x**2)

    def bar():
        x = bar_q.get()
        bar_soln.append(x**3)

    threads = []
    for _ in x_in:
        threads.append(Thread(target=foo))
        threads.append(Thread(target=bar))

    for x in x_in:
        ann.put(x)

    bar_q = ann.subscribe()
    ann.finalize()

    for t in threads:
        t.start()

    for t in threads:
        t.join()

    assert(len(foo_soln) == len(x_in))
    assert(len(bar_soln) == len(x_in))

    for x, r in zip(x_in, foo_soln):
        assert(r == x**2)

    for x, r in zip(x_in, bar_soln):
        assert(r == x**3)


def test_3():
    """ Test subscribing after finalize
    """

    ann = lox.Announcement()
    ann.finalize()
    try:
        ann.subscribe()
        assert(False)  # Subscribe should raise an error
    except lox.SubscribeFinalizedError:
        return
    assert(False)  # Should never reach here


def test_4():
    """ Testing many-to-many capability.
    """

    ann_foo = Announcement()
    ann_bar = ann_foo.subscribe()
    ann_baz = ann_bar.subscribe()

    ann_foo.put(1)

    assert(ann_bar.get() == 1)
    assert(ann_baz.get() == 1)

    ann_bar.put(2)
    assert(ann_foo.get() == 2)
    assert(ann_baz.get() == 2)

    ann_bar.put(3)
    assert(ann_foo.get() == 3)
    assert(ann_baz.get() == 3)


def test_5():
    """ Testing many-to-many backlog and finalization.
    """

    ann_foo = lox.Announcement(backlog=-1)
    ann_bar = ann_foo.subscribe()

    ann_foo.put(1)
    ann_foo.put(3)
    ann_foo.put(7)

    assert(ann_bar.get() == 1)
    assert(ann_bar.get() == 3)
    assert(ann_bar.get() == 7)

    ann_baz = ann_bar.subscribe()

    assert(ann_baz.get() == 1)
    assert(ann_baz.get() == 3)
    assert(ann_baz.get() == 7)

    ann_boo = ann_bar.subscribe()
    ann_boo.finalize()

    assert(ann_boo.get() == 1)
    assert(ann_boo.get() == 3)
    assert(ann_boo.get() == 7)

    try:
        ann_boo.get(timeout=0.01)
        assert(False)  # Above get should raise exception
    except queue.Empty:
        pass

    try:
        ann_foo.subscribe()  # Should raise exception
        assert(False)
    except lox.SubscribeFinalizedError:
        pass



"""
.. module:: process
   :synopsis: Easily execute a function or method in multiple processes.

Still allows the decorated function/method as normal.

Example:

    .. doctest::
        :skipif: True

        >>> import lox
        >>> 
        >>> @lox.process(4) # Will operate with a maximum of 4 processes
        ... def foo(x,y):
        ...     return x*y
        >>> foo(3,4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i+1)
        >>> # foo is currently being executed in 4 processes
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]

"""


from collections import deque
import pathos.multiprocessing as mp
import logging as log
import sys
from .worker import WorkerWrapper
from ..helper import auto_adapt_to_methods, MethodDecoratorAdaptor

__all__ = ['process', ]


class _ProcessWrapper(WorkerWrapper):
    """Process helper decorator.
    """

    def __init__(self, func, n_workers=None):
        if n_workers is None:
            n_workers = mp.cpu_count()
        super().__init__(n_workers, func)
        self.pool = None

    def __len__(self):
        """ 
        Returns
        -------
        int
            Number of jobs not yet completed.
        """

        count = 0
        for res in self.response:
            if not res.ready():
                count += 1

        return count

    def scatter(self, *args, **kwargs):
        """Enqueue a job to be processed by workers.
        Spin up workers if necessary.

        Returns
        -------
        int
            Index into solution list.
        """
        if self.pool is None:
            self.pool = mp.Pool(self.n_workers,)
        self.response.append(self.pool.apply_async(self.func, args=args, kwds=kwargs))
        return len(self.response)-1

    def gather(self):
        """ Gather results. Blocks until job_queue is empty.
        Also blocks scatter on other processes.

        Returns
        -------
        list
            Results in order scatter'd
        """

        self.pool.close()
        self.pool.join()
        del self.pool
        fetched = [x.get() for x in self.response]
        self.response = deque()
        self.pool = None
        return fetched


def process(n_workers):
    """ Decorator to execute a function/method in multiple processes.

    Example:

    .. doctest::
        :skipif: True

        >>> import lox
        >>> 
        >>> @lox.process(4) # Will operate with a maximum of 4 processes
        ... def foo(x,y):
        ...     return x*y
        >>> foo(3,4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i+1)
        >>> # foo is currently being executed in 4 processes
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]

    Parameters
    ----------
    n_workers : int
        Number of process workers to invoke. Defaults to number of CPU cores.

    Methods
    -------
    __call__( *args, **kwargs )
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
            Return of decorated function.

    __len__()
        Returns
        -------
        int
            job queue length.

    scatter( *args, **kwargs )
        Start a job executing ``func( *args, **kwargs )``.
        Workers are created and destroyed automatically.

        Returns
        -------
        int
            Solution's index into the results obtained via ``gather()``.

    gather()
        Block until all jobs called via ``scatter()`` are complete.

        Returns
        -------
        list
            Results in the order that scatter was invoked.
    """

    @auto_adapt_to_methods
    def wrapper(func):
        return _ProcessWrapper(func, n_workers=n_workers)

    if isinstance(n_workers, int):
        # assume this is being called from decorator like "lox.process(5)"
        return wrapper
    else:
        # assume decorator with called as "lox.process"
        func = n_workers
        return MethodDecoratorAdaptor(_ProcessWrapper, func)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
.. module:: helper
   :synopsis: Private general lox helper functions/classes
"""

import sys
import gc


__all__ = ["auto_adapt_to_methods", "MethodDecoratorAdaptor", "term_colors", ]

cdf = {}  # cached (decorator, func)
cdfi = {}  # cached (instance, owner, decorator, func)


def cdfi_gc_cb(phase, info):
    """ Garbage Collector callback to free deleted cdfi entries"""

    global cdfi
    if phase == "start":
        try:
            remove = [k for k, v in cdfi.items() if sys.getrefcount(v.func.__self__) <= 3]
            for k in remove:
                del cdfi[k]
        except RuntimeError:
            pass

gc.callbacks.append(cdfi_gc_cb)


class MethodDecoratorAdaptor:
    """ Class that allows the same decorator apply to methods and functions """

    def __init__(self, decorator, func):
        """
        Parameters
        ----------
        decorator : function
            Decorator function that takes in a single parameter, "func"
        func : function
            Function/Method that is being decorated
        """
        self.decorator = decorator
        self.func = func

    def __call__(self, *args, **kwargs):
        return self._get_cdf()(*args, **kwargs)

    def __get__(self, instance, owner):
        global cdfi
        k = (instance, owner, self.decorator, self.func)
        if k not in cdfi:
            cdfi[k] = self.decorator(self.func.__get__(instance, owner))
        return cdfi[k]

    def __getattr__(self, attr):
        return getattr(self._get_cdf(), attr)

    def __len__(self,):
        return len(self._get_cdf())

    def _get_cdf(self):
        global cdf
        k = (self.decorator, self.func)
        if k not in cdf:
            cdf[k] = self.decorator(self.func)
        return cdf[k]


def auto_adapt_to_methods(decorator):
    """Decorator that allows you to use the same decorator on methods and 
    functions, hiding the self argument from the decorator.

    Source: https://stackoverflow.com/a/1288936"""

    def adapt(func):
        return MethodDecoratorAdaptor(decorator, func)
    return adapt


class term_colors:
    """ Escape sequences to colorize text in the terminal """

    HEADER = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    ORANGE = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    def __init__(self, color: str):
        self.color = color.lower()

    def __enter__(self, ):
        if self.color == 'red':
            print(self.RED, end='')
        elif self.color == 'green':
            print(self.GREEN, end='')
        elif self.color == 'blue':
            print(self.BLUE, end='')
        elif self.color == 'orange' or self.color == 'yellow':
            print(self.ORANGE, end='')
        elif self.color == 'bold':
            print(self.BOLD, end='')
        elif self.color == 'underline':
            print(self.UNDERLINE, end='')
        elif self.color == 'header':
            print(self.HEADER, end='')
        else:
            raise ValueError("Invalid color string")

    def __exit__(self, type, value, traceback):
        print(self.ENDC, end='', flush=True)



from threading import Lock, Thread
from lox import LightSwitch
from copy import copy
from time import sleep, time

SLEEP_TIME = 0.01
N_WORKERS = 5

common_lock = None
counting_lock = None
resp_lock = None
resource = None


def common_setup():
    global common_lock, counting_lock, resp_lock

    common_lock = Lock()
    counting_lock = LightSwitch(common_lock)
    resp_lock = Lock()


def test_LightSwitch_1():
    global common_lock, counting_lock, resp_lock, resource
    common_setup()

    with counting_lock:
        acquired_lock = common_lock.acquire(timeout=0)
        assert(acquired_lock is False)
    acquired_lock = common_lock.acquire(timeout=0)
    assert(acquired_lock is True)
    common_lock.release()


def test_LightSwitch_len():
    global common_lock, counting_lock, resp_lock, resource
    common_setup()
    assert(0 == len(counting_lock))
    counting_lock.acquire()
    assert(1 == len(counting_lock))
    counting_lock.acquire()
    assert(2 == len(counting_lock))
    counting_lock.release()
    assert(1 == len(counting_lock))
    counting_lock.release()
    assert(0 == len(counting_lock))


def test_LightSwitch_timeout():
    lock = Lock()
    ls = LightSwitch(lock)
    lock.acquire()
    assert(ls.acquire(timeout=SLEEP_TIME) is False)
    assert(ls.acquire(timeout=0) is False)


def test_bathroom_example():
    sol = [
        "p_0_enter",
        "p_1_enter",
        "p_0_exit",
        "p_2_enter",
        "p_1_exit",
        "p_3_enter",
        "p_2_exit",
        "p_4_enter",
        "p_3_exit",
        "p_4_exit",
        "j_enter",
        "j_exit",
    ]

    res = bathroom_example()

    for r, s in zip(res, sol):
        assert(r == s)


def bathroom_example():
    """
    Scenario:
        A janitor needs to clean a restroom, but is not allowed to enter until
        all people are out of the restroom. How do we implement this?
    """
    restroom_occupied = Lock()
    restroom = LightSwitch(restroom_occupied)
    res = []
    n_people = 5
    sleep_time = 0.2

    def janitor():
        with restroom_occupied:  # block until the restroom is no longer occupied
            res.append('j_enter')
            print("(%0.3f s) Janitor  entered the restroom" % (time() - t_start,))
            sleep(sleep_time)  # clean the restroom
            res.append('j_exit')
            print("(%0.3f s) Janitor  exited  the restroom" % (time() - t_start,))

    def people(id):
        if id == 0:  # Get the starting time of execution for display purposes
            global t_start
            t_start = time()
        with restroom:  # block if a janitor is in the restroom
            res.append("p_%d_enter" % (id,))
            print("(%0.3f s) Person %d entered the restroom" % (time() - t_start, id,))
            sleep(sleep_time)  # use the restroom
            res.append("p_%d_exit" % (id,))
            print("(%0.3f s) Person %d exited  the restroom" % (time() - t_start, id,))

    people_threads = [Thread(target=people, args=(i,)) for i in range(n_people)]
    janitor_thread = Thread(target=janitor)

    for i, person in enumerate(people_threads):
        person.start()                 # Person i will now attempt to enter the restroom
        sleep(sleep_time * 0.6)        # wait for 60% the time a person spends in the restroom
        if i == 0:                       # While the first person is in the restroom...
            janitor_thread.start()     # the janitor would like to enter. HOWEVER...
            print("(%0.3f s) Janitor Dispatched" % (time()-t_start))
    # Wait for all threads to finish
    for t in people_threads:
        t.join()
    janitor_thread.join()

    # The results will look like:
    """
    Running Restroom Demo
    (0.000 s) Person 0 entered the restroom
    (0.061 s) Person 1 entered the restroom
    (0.100 s) Person 0 exited  the restroom
    (0.122 s) Person 2 entered the restroom
    (0.162 s) Person 1 exited  the restroom
    (0.182 s) Person 3 entered the restroom
    (0.222 s) Person 2 exited  the restroom
    (0.243 s) Person 4 entered the restroom
    (0.282 s) Person 3 exited  the restroom
    (0.343 s) Person 4 exited  the restroom
    (0.343 s) Janitor  entered the restroom
    (0.443 s) Janitor  exited  the restroom
    """
    # Note that multiple people can be in the restroom.
    # If people kept using the restroom, the Janitor would never be able
    # to enter (technically known as thread starvation).
    # If this is undesired for your application, look at RWLock

    return res


if __name__ == "__main__":
    print("Running Restroom Demo")
    bathroom_example()



"""
.. module:: thread
   :synopsis: Easily execute a function or method in multiple threads.

Still allows the decorated function/method as normal.

Example:

.. doctest::

    >>> import lox
    >>> 
    >>> @lox.thread(4) # Will operate with a maximum of 4 threads
    ... def foo(x,y):
    ...     return x*y
    >>> foo(3,4)
    12
    >>> for i in range(5):
    ...     foo.scatter(i, i+1)
    -ignore-
    >>> # foo is currently being executed in 4 threads
    >>> results = foo.gather() # block until results are ready
    >>> print(results) # Results are in the same order as scatter() calls
    [0, 2, 6, 12, 20]
"""

from collections import namedtuple, deque
from threading import Lock, BoundedSemaphore
import logging as log
import queue
import threading
from time import time, sleep
import traceback
import sys
from ..lock import LightSwitch
from ..helper import auto_adapt_to_methods, MethodDecoratorAdaptor, term_colors
from ..queue import Announcement
from .worker import WorkerWrapper, ScatterPromise

__all__ = ["thread", ]


class Job:
    """ Elements on the Job Queue

    Attributes
    ----------
    index
        index of results to store func's return value(s).
    func
        function to execute
    pre_args
        positional arguments to feed into func before promised results (if available)
    post_args
        positional arguments to feed into func after promised results (if available)
    kwargs
        keyword arguments to feed into func
    """

    def __init__(self, index, func, pre_args, post_args, kwargs):
        self.index = index
        self.func = func
        self.pre_args = pre_args
        self.args = ()
        self.post_args = post_args
        self.kwargs = kwargs

    def __str__(self):
        return "(index=%s, func=%s, pre_args=%s, args=%s, post_args=%s, kwargs=%s)" % \
            (str(self.index), str(self.func),
             str(self.pre_args), str(self.args), str(self.post_args), str(self.kwargs))


Result = namedtuple('Result', ['index', 'value', ])


class _ThreadWorker(threading.Thread):
    """Thread worker created on-demand by _ThreadWrapper
    """

    def __init__(self, job_queue, res_queue, res, worker_sem, lightswitch, **kwargs):
        self.job_queue = job_queue   # Queue to pop jobs off of
        self.res_queue = res_queue   # Queue to push results onto
        self.res = res         # deque object to place results in
        self.worker_sem = worker_sem  # object to "release()" upon worker destruction
        self.lightswitch = lightswitch  # object to "release()" upon job completion
        super().__init__(**kwargs)

    def run(self, ):
        """ Executes a Job and stores the results

        Parameters
        ----------
        timeout : float
            Maximum number of seconds to wait while dequeuing a job.
            If no job is dequeued, worker will destroy itself.
        """

        while True:
            try:
                job = self.job_queue.get(timeout=0)
                try:
                    log.debug("Executing decorated function %s with pre_args %s, "
                              "args %s, post_args %s, and kwargs %s" % (
                                  str(job.func), str(job.pre_args),
                                  str(job.args), str(job.post_args),
                                  str(job.kwargs)
                              ))
                    self.res[job.index] = job.func(
                        *job.pre_args, *job.args, *job.post_args, **job.kwargs)
                except Exception:
                    with term_colors("red"):
                        print(traceback.format_exc())
                finally:
                    result = Result(job.index, self.res[job.index])
                    log.debug("Placing %s onto res_queue %s" % (str(result), str(self.res_queue)))
                    self.res_queue.put(result)
                    self.lightswitch.release()
            except queue.Empty:
                # Allow worker to self-terminate
                break
        self.worker_sem.release()  # indicate worker is terminated
        return


class _PromiseForwarder(threading.Thread):
    """ Forwards Promises from the announcement to the job queue. """

    sleep_time = 0.01

    def __init__(self, thread_wrapper, **kwargs):
        """ Create a _PromiseForwarder object

        Parameters
        ----------
        thread_wrapper : _ThreadWrapper
            decorator of current processing function
        """

        super().__init__(**kwargs)
        self.thread_wrapper = thread_wrapper
        self._kill = False

    def run(self,):
        while True:
            if self._kill:
                break

            if self.thread_wrapper.prev_promise is None:
                break

            # Get a result from the previous stage
            log.debug("Trying to get from res_queue %s" %
                      (str(self.thread_wrapper.prev_promise_q),))
            try:
                # Timeout so we can poll the self._kill attribute
                result = self.thread_wrapper.prev_promise_q.get(timeout=1)
            except queue.Empty:
                continue
            log.debug("Popped result %s" % str(result))

            # Get the job
            # Need to loop incase worker threads are faster than main thread
            for i in range(10):
                try:
                    job = self.thread_wrapper.jobs[result.index]
                    break
                except Exception as e:
                    log.warning(e)
                    sleep(self.sleep_time)
            else:
                # Failure
                log.error("Failed to obtain job at %d" % result.index)
                continue
            log.debug("PromiseForwarder: Retrieved job %s" % str(job))

            # Insert the previous results into the new job's args
            if self.thread_wrapper.auto_unpack:
                if isinstance(result.value, tuple):
                    job.args = result.value
                else:
                    job.args = (result.value,)
            else:
                log.debug("Not auto-unpacking")
                job.args = (result.value,)

            log.debug("PromiseForwarder: forwarding job %s" % str(job))
            self.thread_wrapper._dispatch_job(job)
        log.debug("PromiseForwarder Exiting")
        return

    def kill(self):
        """ Sets a flag that the worker periodically checks. When detected,
        worker will destroy itself.
        """

        self._kill = True


class _ThreadWrapper(WorkerWrapper):
    """Thread helper decorator.
    """

    def __init__(self, func, n_workers=50, daemon=None):
        """
        Creates the callable object for the 'thread' decorator.

        Parameters
        ----------
        func : function
            Function handle that each thread worker will execute.

        n_workers : int
            Maximum number of threads to invoke.

        """
        super().__init__(n_workers, func)

        self.job_queue = queue.Queue()  # Queue to put jobs on
        self.jobs_complete_lock = Lock()
        # Used to determine if all jobs have been completed
        self.job_lightswitch = LightSwitch(self.jobs_complete_lock)

        self.workers_sem = BoundedSemaphore(self.n_workers)
        self.daemon = daemon
        self.enable_auto_unpacking()

        self.clear()

    def clear(self):
        log.debug("_ThreadWrapper clearing")
        self.res_queue = Announcement(backlog=0)  # Queue to put results on
        self.promise_forwarder = None  # Thread handle for the promise forwarding task
        self.promises = deque()        # List of promises
        self.jobs = deque()            # list of jobs;
        self.response = deque()
        self.prev_promise = None       # In chaining, a previous chain's promise
        self.prev_promise_q = None

    def disable_auto_unpacking(self):
        """ Automatically unpack previously chained input tuples. 
        Enabled by default.
        """

        self.auto_unpack = False

    def enable_auto_unpacking(self):
        """ Do not unpack previously chained input tuples. """

        self.auto_unpack = True

    def __len__(self):
        """ 
        Returns
        -------
            Length of unprocessed job queue.
        """

        return self.job_queue.qsize()

    def _create_worker(self):
        """ Create a worker if under maximum capacity. """

        if self.workers_sem.acquire(timeout=0):
            _ThreadWorker(self.job_queue, self.res_queue, self.response,
                          self.workers_sem, self.job_lightswitch, daemon=self.daemon).start()

    def _create_promise_forwarder(self):
        """ Create promise forwarding thread. """

        if self.promise_forwarder is None:
            log.debug("Creating promise forwarder thread")
            self.promise_forwarder = _PromiseForwarder(self,)
            self.promise_forwarder.start()

    def _dispatch_job(self, job):
        self.job_queue.put(job)
        self._create_worker()  # Create a thread if we are not at capacity

    def scatter(self, *args, detect_chaining=True, **kwargs):
        """Enqueue a job to be processed by workers.
        Spin up workers if necessary.

        Return
        ------
            Index into the subsequent gather() results
        """
        self.job_lightswitch.acquire()
        index = len(self.response)
        self.response.append(None)

        promise = ScatterPromise(index, self)
        self.promises.append(promise)

        # Detect if these are actual arguments or a single promise (implies chaining)
        prev_promise = None
        if detect_chaining:
            pre_args, post_args = [], []
            for arg in args:
                if isinstance(arg, ScatterPromise):
                    if prev_promise:
                        raise ValueError(
                            "There can only be one promise. "
                            "If your input takes more than one promise, "
                            "you should probably be calling \"gather()\""
                        )
                    else:
                        prev_promise = arg
                    continue
                if prev_promise is not None:
                    post_args.append(arg)
                else:
                    pre_args.append(arg)
            pre_args = tuple(pre_args)
            post_args = tuple(post_args)
        else:
            pre_args = args
            post_args = ()

        job = Job(index, self.func, pre_args, post_args, kwargs)
        self.jobs.append(job)

        if prev_promise is not None:
            if self.prev_promise is None:
                self.prev_promise_q = prev_promise.dec.res_queue.subscribe()
                self.prev_promise = prev_promise
        else:
            self._dispatch_job(job)

        if self.prev_promise is not None:
            self._create_promise_forwarder()

        return promise

    def _finalize(self):
        """ Finalize self.res_queue and previous chain. """

        self.res_queue.finalize()
        if self.prev_promise is not None:
            self.prev_promise.dec._finalize()

    def gather(self):
        """ Block and collect results from prior ``scatter`` calls.
        """

        log.debug("Gathering %s" % (str(self.func),))

        # Save res_queue backlog memory
        self._finalize()

        with self.jobs_complete_lock:  # Wait until all jobs are done
            log.debug("Gathering %s: jobs_complete_lock acquired" % (str(self.func)))
            response = list(self.response)

            if self.promise_forwarder is not None:
                log.debug("Gathering previous %s" % (str(self.prev_promise.dec.func),))

                self.promise_forwarder.kill()
                self.promise_forwarder = None
                self.prev_promise.dec.gather()

            # Clear internal results
            self.clear()

        return response


def thread(max_workers, daemon=None):
    """ Decorator to execute a function in multiple threads.

    Example:

    .. doctest::

        >>> import lox
        >>> 
        >>> @lox.thread(4) # Will operate with a maximum of 4 threads
        ... def foo(x,y):
        ...     return x*y
        >>> foo(3,4)
        12
        >>> for i in range(5):
        ...     foo.scatter(i, i+1)
        -ignore-
        >>> # foo is currently being executed in 4 threads
        >>> results = foo.gather()
        >>> print(results)
        [0, 2, 6, 12, 20]

    Multiple decorated functions can be chained together, each function drawing
    from their own pool of threads. Functions that return tuples will automatically
    unpack into the chained function. Positional arguments and keyword arguments
    can be passed in as they normally would.

    .. doctest::
        >>> @lox.thread(2) # Will operate with a maximum of 2 threads
        ... def bar(x,y):
        ...     return x + y

        >>>  for i in range(5):
        ...    foo_res = foo.scatter(i, i+1)
        ...    bar.scatter(foo_res, 10) # scatter will automatically unpack the results of foo 
        >>>  
        >>> results = bar.gather() 

    Currently, a ``scatter`` call can have a maximum of 1 previous ``scatter``
    result as an input argument. However, unlimited number of functions can 
    be chained together in any topology.

    Parameters
    ----------
    max_workers : int
        Maximum number of threads to invoke.
        When ``lox.thread`` is called without ``()``, the wrapped function 
        a default number of max_workers is used (50).

    Methods
    -------
    __call__( *args, **kwargs )
        Vanilla passthrough function execution. Default user function behavior.

        Returns
        -------
        Decorated function return type.
           Return of decorated function.

    __len__()
        Returns
        -------
        int
            Current job queue length. Number of jobs that are currently waiting
            for an available worker.

    scatter( *args, **kwargs)
        Start a job executing decorated function ``func( *args, **kwargs )``.
        Workers are created and destroyed automatically.

        Returns
        -------
        int
            Solution's index into the results obtained via ``gather()``.

    gather()
        Block until all jobs called via ``scatter()`` are complete.

        Returns
        -------
        list
            Results in the order that scatter was invoked.

    disable_auto_unpacking()
        Automatically unpack previously chained input tuples.

    enable_auto_unpacking()
        Do not unpack previously chained input tuples.

    """

    @auto_adapt_to_methods
    def wrapper(func):
        return _ThreadWrapper(func, n_workers=max_workers, daemon=daemon)

    if isinstance(max_workers, int):
        # assume this is being called from decorator like "lox.thread(5)"
        return wrapper
    else:
        # assume decorator with called as "lox.thread"
        func = max_workers
        return MethodDecoratorAdaptor(_ThreadWrapper, func)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



import lox
from time import sleep, time

SLEEP_TIME = 0.01
N_WORKERS = 2


def test_basic_args():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.process(N_WORKERS)  # specifying maximum number of processes
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert(10 == worker_task(2, 5))

    assert(len(worker_task) == 0)
    print("test_process: %s" % (worker_task))

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)

    res = worker_task.gather()
    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y) == r)


def test_basic_noargs():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]

    @lox.process  # maximum number of processes = # cores
    def worker_task(x, y):
        sleep(SLEEP_TIME)
        return x * y

    # Vanilla function execution still works
    assert(10 == worker_task(2, 5))

    assert(len(worker_task) == 0)
    print("test_process: %s" % (worker_task))

    for x, y in zip(in_x, in_y):
        worker_task.scatter(x, y)

    res = worker_task.gather()
    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y) == r)


class Class1:

    def __init__(self, z):
        self.z = z

    @lox.process(2)
    def test_method1(self, x, y):
        sleep(SLEEP_TIME)
        return x*y + self.z

    @lox.process
    def test_method2(self, x, y):
        sleep(SLEEP_TIME)
        return x*y + self.z


def test_method_1():
    in_x = [1,   2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, ]
    in_y = [13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, ]
    z = 5

    test_obj = Class1(z)

    assert((2*5+z) == test_obj.test_method1(2, 5))
    assert((2*5+z) == test_obj.test_method2(2, 5))

    for x, y in zip(in_x, in_y):
        test_obj.test_method1.scatter(x, y)
    res = test_obj.test_method1.gather()
    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y+z) == r)

    for x, y in zip(in_x, in_y):
        test_obj.test_method2.scatter(x, y)
    res = test_obj.test_method2.gather()
    assert(len(res) == len(in_x))

    for r, x, y in zip(res, in_x, in_y):
        assert((x*y+z) == r)



"""
.. module:: funnel
   :synopsis: Wait on many queues.

Wait on a queue until a set of inputs are ready.
"""


import queue
from collections import deque
import threading
import logging as log

__all__ = ["Funnel", "FunnelPutError", "FunnelPutTopError", ]


class FunnelPutError(Exception):
    """ Cannot ``put`` to a top/master Funnel object;
    can only ``put`` to subscribers.
    """

    pass


class FunnelPutTopError(Exception):
    """ Can only put onto subscribers, not the top/master ``Funnel``.
    """

    pass


class FunnelElement:

    def __init__(self, item=None, complete=False):
        """
        Parameters
        ----------
        item
            User data
        complete : bool
            True if item contains valid data. 
        """

        self.item = item
        self.complete = complete

    def set(self, item):
        self.item = item
        self.complete = True


class Funnel:
    """ Wait on many queues.

    .. doctest::

        >>> funnel = lox.Funnel()
        >>> sub_1 = funnel.subscribe()
        >>> sub_2 = funnel.subscribe()

        >>> sub_1.put('foo', 'job_id')
        >>> try:
        ...     res = funnel.get(timeout=0.01)
        ... except queue.Empty:
        ...     print("Timed Out")
        Timed Out
        >>> sub_2.put('bar', 'job_id')
        >>> res = funnel.get()
        >>> print(len(res))
        3
        >>> print(res)
        ['job_id','foo','bar']


    Attributes
    ----------
    index : int
        Index into list of solutions (if a subscriber). ``-1`` otherwise.
        Note: if ``get(return_jid=True)`` then this is offset by one.
    """

    def __init__(self):
        """ Create a Funnel Object
        """

        self.lock = threading.RLock()
        self.index = -1  # Index into list of solutions
        self.subscribers = deque()
        self.d = {}
        self.q = queue.Queue()          # Object that complete sets are placed on.

    @classmethod
    def _clone(cls, funnel):
        """ Create a new ``Funnel`` object that shares subscribers and resources
        with an existing ``Funnel``.

        Only difference is it's index is 1 higher than the funnel's last subscriber.

        Parameters
        ----------
        funnel : lox.Funnel
            ``Funnel`` object to clone from

        Returns
        -------
        Funnel
           New Funnel object.
        """

        new_funnel = cls()
        new_funnel.lock = funnel.lock

        with new_funnel.lock:
            new_funnel.d = funnel.d
            new_funnel.q = funnel.q
            new_funnel.subscribers = funnel.subscribers
            new_funnel.index = len(funnel)
            log.debug("new_funnel.index == %d" % new_funnel.index)
            new_funnel.subscribers.append(new_funnel)

        return new_funnel

    def __len__(self,):
        """ Gets number of input queues.

        Returns
        -------
        int
            Number of input queues.
        """

        return len(self.subscribers)

    def subscribe(self,):
        """ Create a new Funnel for data to be ``put`` on. 

        Returns
        -------
        Funnel
            A funnel object that is a required input on ``get`` calls.
        """

        new_funnel = Funnel._clone(self)
        return new_funnel

    def put(self, item, jid, blocking=True, timeout=-1, ):
        """
        Parameters
        ----------
        item
           data to put onto all subscribers' queues

        jid : hashable
           unique identifier for job.

        block : bool
            Block until data is put on queues or timeout.

        timeout : float
            Wait up to ``timeout`` seconds before raising ``queue.Full``.
            Defaults to no timeout.

        Returns
        -------
        bool
            True if item was successfully added; False otherwise.

        Raises
        ------
        FunnelPutTopError
           Can only put onto subscribers, not the top/master ``Funnel``.
        """

        if self.index < 0:
            raise FunnelPutTopError

        if not self.lock.acquire(blocking=blocking, timeout=timeout):
            return False

        # put Item into dict's key's deque
        val = self.d.get(jid)
        if val is None:
            self.d[jid] = deque()

        for _ in range(len(self) - len(self.d[jid])):
            # Add a empty, incomplete funnel elements until it matches
            # the number of input queues.
            self.d[jid].append(FunnelElement())

        self.d[jid][self.index].set(item)

        # Add item to queue if all subscribers are accounted for.
        if all([elem.complete for elem in self.d[jid]]):
            log.debug("JID \"%s\" complete; putting onto queue %s" % (str(jid), str(self.q)))
            self.q.put(tuple([jid, ] + [elem.item for elem in self.d[jid]]))
            del self.d[jid]

        self.lock.release()
        return True

    def get(self, block=True, timeout=None, return_jid=True):
        """ Get from the receive queue. Will return the contents of each 
        input queue in the order subscribed as a tuple

        Parameters
        ----------
        block : bool
            Block until data is obtained from receive queue or timeout.

        timeout : float
            Wait up to ``timeout`` seconds before raising ``queue.Full``.
            Defaults to no timeout.

        return_jid : bool
            Have the Job ID as the first element of the returned tuple.
            Defaults to True

        Returns
        -------
        tuple
            items from input queues.

        Raises
        ------
        queue.Empty
             When there are no elements in queue and timeout has been reached.
        """
        items = self.q.get(block=block, timeout=timeout)

        if return_jid:
            return items
        else:
            return items[1:]



# -*- coding: utf-8 -*-

"""
Easy multithreading for every project.
"""

__author__ = """Brian Pugh"""
__email__ = 'bnp117@gmail.com'
__version__ = '0.6.3'

from .lock import *
from .queue import *
from .worker import *
from .exceptions import *



import lox
from lox import Announcement
from threading import Lock, Thread
from time import sleep
import queue


def test_1():
    """ Test most basic usecase with 1 subscribers.
    """
    funnel = lox.Funnel()
    sub_1 = funnel.subscribe()

    sub_1.put('foo', 'job_id')
    res = funnel.get()
    assert(len(res) == 2)
    assert(res[0] == 'job_id')
    assert(res[1] == 'foo')
    assert(funnel.d == {})


def test_2():
    """ Test most basic usecase with 2 subscribers.
    """

    funnel = lox.Funnel()
    sub_1 = funnel.subscribe()
    sub_2 = funnel.subscribe()

    sub_1.put('foo', 'job_id')
    try:
        res = funnel.get(timeout=0.01)
        assert(0)
    except queue.Empty:
        assert(1)

    sub_2.put('bar', 'job_id')

    res = funnel.get()
    assert(len(res) == 3)
    assert(res[0] == 'job_id')
    assert(res[1] == 'foo')
    assert(res[2] == 'bar')
    assert(funnel.d == {})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
import sys
from setuptools import setup, find_packages
from setuptools.command.install import install

version='0.6.3'


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version

    https://circleci.com/blog/continuously-deploying-python-packages-to-pypi-with-circleci/
    """

    description = 'verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != 'v'+version:
            info = "Git tag: \"{0}\" does not match the version of this app: \"{1}\"".format(
                tag, version
            )
            sys.exit(info)


with open('README.rst') as readme_file:
    readme = readme_file.read()
readme = readme.replace(
        'assets/lox_200w.png', 
        'https://raw.githubusercontent.com/BrianPugh/lox/master/assets/lox_200w.png',
        )

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', ]

setup(
    author="Brian Pugh",
    author_email='bnp117@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="Threading and Multiprocessing for every project.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/x-rst',
    include_package_data=True,
    keywords='lox',
    name='lox',
    packages=find_packages(include=['lox']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/BrianPugh/lox',
    version=version,
    zip_safe=False,
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)



from threading import Lock, Thread
from lox import IndexSemaphore
from time import time, sleep
from collections import deque

SLEEP_TIME = 0.01
n_resource = 5
n_threads = 20


def test_multithread_args():
    resp = deque()
    sem = IndexSemaphore(n_resource)
    locks = [Lock() for _ in range(n_resource)]

    def func():
        nonlocal locks
        index = sem.acquire(timeout=100)
        if locks[index].acquire(timeout=0):
            # Acquired Test Lock with no waiting, indicating this index is unique
            sleep(SLEEP_TIME)
            locks[index].release()
            resp.append(True)
        else:
            # timeout (bad)
            resp.append(False)
        sem.release(index)

    threads = [Thread(target=func) for _ in range(n_threads)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()
    for r in resp:
        assert r
    assert(len(resp) == n_threads)


def test_multithread_no_args():
    resp = deque()
    sem = IndexSemaphore(n_resource)
    locks = [Lock() for _ in range(n_resource)]

    def func():
        nonlocal locks
        index = sem.acquire()
        if locks[index].acquire(timeout=0):
            # Acquired Test Lock with no waiting, indicating this index is unique
            sleep(SLEEP_TIME)
            locks[index].release()
            resp.append(True)
        else:
            # timeout (bad)
            resp.append(False)
        sem.release(index)

    threads = [Thread(target=func) for _ in range(n_threads)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()
    for r in resp:
        assert r
    assert(len(resp) == n_threads)


def test_multithread_context_args():
    resp = deque()
    sem = IndexSemaphore(n_resource)
    locks = [Lock() for _ in range(n_resource)]

    def func():
        nonlocal locks
        with sem(timeout=None) as index:
            if locks[index].acquire(timeout=0):
                # Acquired Test Lock with no waiting, indicating this index is unique
                sleep(SLEEP_TIME)
                locks[index].release()
                resp.append(True)
            else:
                # timeout (bad)
                resp.append(False)

    threads = [Thread(target=func) for _ in range(n_threads)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()
    for r in resp:
        assert(r)
    assert(len(resp) == n_threads)


def test_multithread_context_no_args():
    resp = deque()
    sem = IndexSemaphore(n_resource)
    locks = [Lock() for _ in range(n_resource)]

    def func():
        nonlocal locks
        with sem() as index:
            if locks[index].acquire(timeout=0):
                # Acquired Test Lock with no waiting, indicating this index is unique
                sleep(SLEEP_TIME)
                locks[index].release()
                resp.append(True)
            else:
                # timeout (bad)
                resp.append(False)

    threads = [Thread(target=func) for _ in range(n_threads)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()
    for r in resp:
        assert r
    assert(len(resp) == n_threads)


def test_invalid_constructor():
    try:
        IndexSemaphore(0)
        assert(False)
    except ValueError:
        assert(True)


def test_timeout():
    sem = IndexSemaphore(1)
    with sem(timeout=None) as index1:
        assert(index1 == 0)
        with sem(timeout=0.1) as index2:
            assert index2 is None



"""
.. module:: announcement
   :synopsis: Many-to-many queue recipients with backlog support for race 
   condition subscribing or lazy subscribing.
"""


import queue
from collections import deque
import threading
import logging as log

__all__ = ["Announcement", "SubscribeFinalizedError", ]


class SubscribeFinalizedError(Exception):
    """ Raised when an Announcement.subscribe has been called,
    but the Announcement has already been finalized.
    """

    pass


class Announcement:
    """ Push to many queues with backlog support.

    Allows the pushing of data to many threads.

    Example:

    .. doctest::

        >>> import lox
        >>> ann = lox.Announcement()
        >>> foo_q = ann.subscribe()
        >>> bar_q = ann.subscribe()
        >>> 
        >>> @lox.thread
        ... def foo():
        ...     x = foo_q.get()
        ...     return x
        >>> 
        >>> @lox.thread
        ... def bar():
        ...     x = bar_q.get()
        ...     return x**2
        >>> 
        >>> ann.put(5)
        >>> foo.scatter()
        >>> foo_res = foo.gather()
        >>> bar.scatter()
        >>> bar_res = bar.gather()

    The backlog allows future (or potentially race-condition) subscribers to 
    get content put'd before they subscribed. However, the user must be careful 
    of memory consumption.

    Attributes
    ----------
    backlog : deque
        Backlog of queued data. None if not used.
    """

    @property
    def final(self,):
        return self._final

    @final.setter
    def final(self, val: bool):
        with self.lock:
            self._final = val

    @property
    def backlog(self,):
        return self._backlog

    @backlog.setter
    def backlog(self, val):
        self._backlog = val

    def __repr__(self,):
        with self.lock:
            qs = [str(id(ann.q)) for ann in self.subscribers]
            string = super().__repr__() + ' with subscriber queues ' + ','.join(qs)
        return string

    def __init__(self, maxsize=0, backlog=None):
        """ Create an Announcement object.

        Parameters
        ----------
        maxsize : int
            Created queue's default maximum size. Defaults to 0 (no size limit).

        backlog : int
            Create a log of previously put data. New subscribers get the entire 
            backlog pushed onto their queue. Memory usage potentially unbounded.

            A value <=0 specifies unbounded backlog size. A value >0 specifies 
            a circular buffer.

            Defaults to None (backlog not used).

            Note: If used, it is recommended to use the ``finalize()`` method so
            that the backlog can be free'd when no longer necessary.
        """

        log.debug("Creating Announcement %X" % id(self))

        # General Attributes #
        self.subscribers = deque()
        self.maxsize = maxsize
        self.lock = threading.RLock()
        self.q = queue.Queue(maxsize=self.maxsize)
        self.subscribers.append(self)

        # Backlog objects #
        self.backlog_use = False if backlog is None else True
        self.final = False
        if self.backlog_use:
            if backlog <= 0:
                backlog = None
            self.backlog = deque(maxlen=backlog)
        else:
            self.backlog = None

    @classmethod
    def clone(cls, ann, q: queue.Queue = None):
        """ Create a new announcement object that shares subscribers and resources
        with an existing announcement.

        Only difference from cloned announcement is a new receive queue is created.

        Parameters
        ----------
        ann : lox.Announcement
            Announcement object to clone from

        q : queue.Queue
            Receiving queue. If ``None``, a new one is created.

        Returns
        -------
        Announcement
           New Announcement object with copied attributes, but new ``q``
        """

        new_ann = cls()
        new_ann.lock = ann.lock
        with new_ann.lock:
            new_ann.subscribers = ann.subscribers
            new_ann.maxsize = ann.maxsize
            new_ann.q = queue.Queue(maxsize=ann.maxsize) if q is None else q
            new_ann.backlog_use = ann.backlog_use
            new_ann.backlog = ann.backlog
        return new_ann

    def __len__(self,):
        """ Get the number of subscribers.

        Returns
        -------
        int
            Number of subcribers.
        """

        return len(self.subscribers)

    def qsize(self,):
        """ Return the approximate size of the receive queue. Note, qsize() > 0 
        doesn’t guarantee that a subsequent get() will not block, nor will 
        qsize() < maxsize guarantee that put() will not block.

        Returns
        -------
        int
            approximate size of the receive queue.
        """

        return self.q.qsize()

    def empty(self,):
        """ Return ``True`` if the receive queue is empty, ``False`` otherwise. 
        If ``empty()`` returns ``True`` it doesn’t guarantee that a subsequent 
        call to ``put()`` will not block. Similarly, if ``empty()`` returns 
        ``False`` it doesn’t guarantee that a subsequent call to ``get()`` 
        will not block.

        Returns
        -------
        bool
            ``True`` if the receive queue is currently empty; ``False`` otherwise.
        """

        return self.q.empty()

    def full(self,):
        """ Return ``True`` if the receive queue is full, ``False`` otherwise. 
        If ``full()`` returns ``True`` it doesn’t guarantee that a subsequent 
        call to ``get()`` will not block. Similarly, if ``full()`` returns 
        ``False`` it doesn’t guarantee that a subsequent call to ``put()`` 
        will not block.

        Returns
        -------
        bool
            ``True`` if the receive queue is currently full; ``False`` otherwise.
        """

        return self.q.full()

    def put(self, item, block=True, timeout=None, ):
        """ Put item into all subscribers' queues. 

        Parameters
        ----------
        item
           data to put onto all subscribers' queues

        block : bool
            Block until data is put on queues or timeout.

        timeout : float
            Wait up to ``timeout`` seconds before raising ``queue.Full``.
            Defaults to no timeout.
        """

        with self.lock:
            for ann in self.subscribers:
                if ann == self:
                    log.debug("Skipping receiver Announcement %s" % ann)
                    continue
                else:
                    log.debug("Putting to Announcement %s" % str(ann))

                ann.q.put(item, block=block, timeout=timeout)
            if self.backlog_use and not self.final:
                self.backlog.append(item)

    def get(self, block=True, timeout=None):
        """ Get from the receive queue. 

        Parameters
        ----------
        block : bool
            Block until data is obtained from receive queue or timeout.

        timeout : float
            Wait up to ``timeout`` seconds before raising ``queue.Full``.
            Defaults to no timeout.

        Returns
        -------
            item from receive queue.

        Raises
        ------
        queue.Empty
             When there are no elements in queue and timeout has been reached.
        """

        return self.q.get(block=block, timeout=timeout)

    def subscribe(self, q=None, maxsize=None, block=True, timeout=None):
        """ Subscribe to announcements.

        Parameters
        ----------
        q : Queue
            Existing queue to add to the subscribers' list.
            If not provided, a queue is created.

        maxsize : int
            Created queue's maximum size. Overrides Announcement's default 
            maximum size. Ignored if ``q`` is provided.

        block : bool
            Block until data from backlog is put on queues or timeout.

        timeout : float
            Wait up to ``timeout`` seconds before raising ``queue.Full``.
            Defaults to no timeout.

        Returns
        -------
        Announcement
            object for receiver to ``get`` and ``put`` data from.
        """

        with self.lock:
            if self.final:
                log.error("%s Subscribe when announcement already finalized" % (self,))
                raise SubscribeFinalizedError

            if q is None:
                # Create Queue
                if maxsize is None:
                    maxsize = self.maxsize
                q = queue.Queue(maxsize=maxsize)
            ann = Announcement.clone(self, q)
            self.subscribers.append(ann)
            if self.backlog_use:
                for x in self.backlog:
                    q.put(x, block=block, timeout=timeout)
        return ann

    def unsubscribe(self, q):
        """ Remove the queue from queue-list. Will no longer receive announcements.

        Parameters
        ----------
        q : Queue
            Queue object to remove.

        Raises
        ------
        ValueError
            ``q`` was not a subscriber.
        """

        with self.lock:
            self.subscribers.remove(q)

    def finalize(self,):
        """ Do not allow any more subscribers.

        Primarily used for memory efficiency if backlog is used.
        """

        with self.lock:
            for ann in self.subscribers:
                log.debug("%s finalizing" % str(ann))
                ann.final = True
                ann.backlog = None
